# Copyright 2021 Patrick C. Tapping
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Error code definitions and exception class for dealing with errors reported from the Andor SDK.
"""

__all__ = ["AndorError", "AT_ERR"]

from enum import IntEnum

class AndorError(RuntimeError):
    """
    Exception used to indicate an error returned by the Andor SDK3.

    :param error_code: Numerical error code used by the Andor SDK3.
    """

    def __init__(self, error_code):
        self.code = error_code
        self.name = AT_ERR(self.code).name
        self.description = AT_ERR(error_code).description
    
    def __str__(self):
        return f"AT_ERR_{self.name} ({self.code}): {self.description}"


class AT_ERR(IntEnum):
    """
    Enumeration listing the valid Andor SDK3 numerical error codes.
    """
    SUCCESS = 0
    NOTINITIALISED = 1
    NOTIMPLEMENTED = 2
    READONLY = 3
    NOTREADABLE = 4
    NOTWRITABLE = 5
    OUTOFRANGE = 6
    INDEXNOTAVAILABLE = 7
    INDEXNOTIMPLEMENTED = 8
    EXCEEDEDMAXSTRINGLENGTH = 9
    CONNECTION = 10
    NODATA = 11
    INVALIDHANDLE = 12
    TIMEDOUT = 13
    BUFFERFULL = 14
    INVALIDSIZE = 15
    INVALIDALIGNMENT = 16
    COMM = 17
    STRINGNOTAVAILABLE = 18
    STRINGNOTIMPLEMENTED = 19
    NULL_FEATURE = 20
    NULL_HANDLE = 21
    NULL_IMPLEMENTED_VAR = 22
    NULL_READABLE_VAR = 23
    NULL_READONLY_VAR = 24
    NULL_WRITABLE_VAR = 25
    NULL_MINVALUE = 26
    NULL_MAXVALUE = 27
    NULL_VALUE = 28
    NULL_STRING = 29
    NULL_COUNT_VAR = 30
    NULL_ISAVAILABLE_VAR = 31
    NULL_MAXSTRINGLENGTH = 32
    NULL_EVCALLBACK = 33
    NULL_QUEUE_PTR = 34
    NULL_WAIT_PTR = 35
    NULL_PTRSIZE = 36
    NOMEMORY = 37
    DEVICEINUSE = 38
    DEVICENOTFOUND = 39
    HARDWARE_OVERFLOW = 100

    @property
    def description(self):
        """
        Return a string describing the error code.
        """
        return {
            AT_ERR.SUCCESS : "Function call was successful",
            AT_ERR.NOTINITIALISED : "Function called with an uninitialized handle",
            AT_ERR.NOTIMPLEMENTED : "Feature has not been implemented for the chosen camera",
            AT_ERR.READONLY : "Feature is read only",
            AT_ERR.NOTREADABLE : "Feature is currently not readable",
            AT_ERR.NOTWRITABLE : "Feature is currently not writable",
            AT_ERR.OUTOFRANGE : "Value is outside the maximum and minimum limits",
            AT_ERR.INDEXNOTAVAILABLE : "Index is currently not available",
            AT_ERR.INDEXNOTIMPLEMENTED : "Index is not implemented for the chosen camera",
            AT_ERR.EXCEEDEDMAXSTRINGLENGTH : "String value provided exceeds the maximum allowed length",
            AT_ERR.CONNECTION : "Error connecting to or disconnecting from hardware",
            AT_ERR.NODATA : "No Internal Event or Internal Error",
            AT_ERR.INVALIDHANDLE : "Invalid device handle passed to function",
            AT_ERR.TIMEDOUT : "The waitBuffer function timed out while waiting for data arrive in output queue",
            AT_ERR.BUFFERFULL : "The input queue has reached its capacity",
            AT_ERR.INVALIDSIZE : "The size of a queued buffer did not match the frame size",
            AT_ERR.INVALIDALIGNMENT : "A queued buffer was not aligned on an 8-byte boundary",
            AT_ERR.COMM : "An error has occurred while communicating with hardware",
            AT_ERR.STRINGNOTAVAILABLE : "Index / String is not available",
            AT_ERR.STRINGNOTIMPLEMENTED : "Index / String is not implemented for the chosen camera",
            AT_ERR.NULL_FEATURE : "NULL feature name passed to function",
            AT_ERR.NULL_HANDLE : "Null device handle passed to function",
            AT_ERR.NULL_IMPLEMENTED_VAR : "Feature not implemented",
            AT_ERR.NULL_READABLE_VAR : "Readable not set",
            AT_ERR.NULL_READONLY_VAR : "Read-only",
            AT_ERR.NULL_WRITABLE_VAR : "Writable not set",
            AT_ERR.NULL_MINVALUE : "NULL min value",
            AT_ERR.NULL_MAXVALUE : "NULL max value",
            AT_ERR.NULL_VALUE : "NULL value returned from function",
            AT_ERR.NULL_STRING : "NULL string returned from function",
            AT_ERR.NULL_COUNT_VAR : "NULL feature count",
            AT_ERR.NULL_ISAVAILABLE_VAR : "Available not set",
            AT_ERR.NULL_MAXSTRINGLENGTH : "Max string length is NULL",
            AT_ERR.NULL_EVCALLBACK : "EvCallBack parameter is NULL",
            AT_ERR.NULL_QUEUE_PTR : "Pointer to queue is NULL",
            AT_ERR.NULL_WAIT_PTR : "Wait pointer is NULL",
            AT_ERR.NULL_PTRSIZE : "Pointer size is NULL",
            AT_ERR.NOMEMORY : "No memory has been allocated for the current action",
            AT_ERR.DEVICEINUSE : "Function failed to connect to a device because it is already being used",
            AT_ERR.DEVICENOTFOUND : "Device not found",
            AT_ERR.HARDWARE_OVERFLOW : "The software was not able to retrieve data from the card or camera fast enough to avoid the internal hardware buffer bursting."
        }[self.value]



# Copyright 2021 Patrick C. Tapping
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["Andor3"]

import os
import platform
import typing
import logging
from ctypes import cdll, byref, c_int, c_longlong, c_double, c_wchar_p, c_void_p, POINTER, CFUNCTYPE, create_unicode_buffer
from ctypes.util import find_library

import numpy as np

from . constants import *
from . error import *
from . utils import *


_log = logging.getLogger(__name__)
"""Logging output for use by this module."""

# ctypes callback type (return type, argument types...)
_CALLBACK_FUNC = CFUNCTYPE(c_int, c_int, c_wchar_p, POINTER(c_longlong))

class Andor3():
    """
    Initialise the Andor SDK3 and attempt to open a connection to a camera.

    If ``device_index`` is None, camera initialisation won't be attempted, however the number of
    cameras attached may be queried with :data:`~Andor3.device_count`, and a camera subsequently opened
    with :meth:`open`.
    
    :param device_index: Index (zero-based) of attached camera to open.
    """

    def __init__(self, device_index:int=0):

        self.camera_handle = AT_HANDLE.UNINITIALISED
        """The "camera handle" used by the SDK to identify the currently opened camera."""

        self.software_version = ""
        """Version number of the SDK3 library in use."""

        self.device_count = 0
        """Number of detected cameras attached to the system."""

        # Andor3 library only exists for Linux or Windows
        self._dll = None
        """Reference to the python ctypes library in use."""
        if platform.system() == "Windows":
            libname = "atcore.dll"
        elif platform.system() == "Linux":
            libname = "libatcore.so.3"
        else:
            raise EnvironmentError("Did not detect a Windows or Linux operating system!")

        # Find and load the Andor library
        dll_path = None
        try:
            # Use system installed library if available
            _log.debug(f"Searching for Andor3 library in system paths...")
            self._dll = cdll.LoadLibrary(libname)
            dll_path = os.path.dirname(find_library(libname))
        except:
            pass
        
        # If that didn't work, try some other common locations
        if self._dll is None and platform.system() == "Windows":
            for loc in (os.path.join(os.environ["ProgramFiles"], "Andor Driver Pack 3"),
                        os.path.join(os.environ["ProgramFiles"], "Andor SOLIS"),
                        os.path.join(os.environ["ProgramFiles"], "National Instruments", "LabVIEW")):    
                try:
                    _log.debug(f"Searching for Andor3 library in {loc}...")
                    self._dll = cdll.LoadLibrary(os.path.join(loc, libname))
                    # If that worked, add location to path so atcore.dll can find the other dlls it needs.
                    os.environ["PATH"] = f"{os.environ['PATH']}{os.pathsep}{loc}"
                    dll_path = loc
                    break
                except:
                    pass
        
        # Check current directory
        if self._dll is None:
            try:
                loc = os.path.abspath(".")
                _log.debug(f"Searching for Andor3 library in {loc}...")
                self._dll = cdll.LoadLibrary(os.path.join(loc, libname))
                # Don't need to add to paths on Windows (atcore.dll must search in current dir)
                # On Linux, maybe need to add to LD_LIBRARY_PATH or something?
                dll_path = loc
            except:
                pass
        
        # Hopefully we found the library somewhere!
        if self._dll is None:
            raise RuntimeError(f"Couldn't find the Andor3 shared library '{libname}'")
        _log.info(f"Initialising Andor3 library found at {dll_path}{os.sep}{libname}...")
        
        # Initialise library
        error = self._dll.AT_InitialiseLibrary()
        if not error == AT_ERR.SUCCESS:
            raise AndorError(error)
        
        # Get SDK version number
        software_version = create_unicode_buffer(STRING_BUFFER_SIZE)
        error = self._dll.AT_GetString(AT_HANDLE.SYSTEM, "SoftwareVersion", byref(software_version))
        if not error == AT_ERR.SUCCESS:
            self._dll.AT_FinaliseLibrary()
            raise AndorError(error)
        self.software_version = software_version.value
        _log.debug(f"Andor3 library version {self.software_version}")


        # Get device count
        if device_index is None: device_index = -1
        device_count = c_int()
        error = self._dll.AT_GetInt(AT_HANDLE.SYSTEM, "DeviceCount", byref(device_count))
        if not error == AT_ERR.SUCCESS:
            self._dll.AT_FinaliseLibrary()
            raise AndorError(error)
        if device_count.value < 1 or device_index + 1 > device_count.value:
            # Raise error if no cameras found, or not enough cameras
            self._dll.AT_FinaliseLibrary()
            raise AndorError(AT_ERR.DEVICENOTFOUND)
        self.device_count = device_count.value
        _log.debug(f"Detected {self.device_count} device{'s' if self.device_count > 1 else ''}")

        # If a device_index is given, attempt to open the camera
        if device_index >= 0:
            try:
                self.open(device_index)
            except:
                self._dll.AT_FinaliseLibrary()
                raise
        
        # Our own record of registered feature callbacks
        self._callbacks = dict()

        # Block of memory for image buffer(s)
        self._image_buffer = None
        # Properties of images at the time the image buffer was created
        self._image_properties = None

    def open(self, device_index:int=0) -> None:
        """
        Open an attached camera device by its numerical device index.

        :param device_index: Index (zero-based) of the camera to open.
        """
        # Close any currently open camera
        self.close()
        # Attempt open of new camera
        _log.debug(f"Opening connection to camera #{device_index}...")
        camera_handle = c_int()
        error = self._dll.AT_Open(device_index, byref(camera_handle))
        if not error == AT_ERR.SUCCESS:
            raise AndorError(error)
        self.camera_handle = camera_handle.value
        _log.debug(f"Camera assigned handle {self.camera_handle}")


    def close(self) -> None:
        """
        Close any currently opened camera.
        """
        if not self.camera_handle == AT_HANDLE.UNINITIALISED:
            error = self._dll.AT_Close(self.camera_handle)
            if not error == AT_ERR.SUCCESS:
                raise AndorError(error)
            self.camera_handle = AT_HANDLE.UNINITIALISED
            _log.debug("Closed connection to camera")

    def __del__(self):
        try:
            self._dll.AT_Close(self.camera_handle)
        except: pass
        try:
            self._dll.AT_FinaliseLibrary()
        except: pass

    def isImplemented(self, feature: str) -> bool:
        """
        Check if the given feature is implemented on the currently opened camera.

        See the :data:`~andor3.constants.FEATURES` dictionary for possible ``feature`` values.
        
        :param feature: String describing the camera feature.
        :returns: ``True`` if feature is implemented, or ``False`` otherwise.
        """
        result = c_int()
        error = self._dll.AT_IsImplemented(self.camera_handle, feature, byref(result))
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
        return bool(result.value)
    
    def isReadable(self, feature: str) -> bool:
        """
        Check if the given feature is currently readable.

        See the :data:`~andor3.constants.FEATURES` dictionary for possible ``feature`` values.
        
        :param feature: String describing the camera feature.
        :returns: ``True`` if feature is readable, or ``False`` otherwise.
        """
        result = c_int()
        error = self._dll.AT_IsReadable(self.camera_handle, feature, byref(result))
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
        return bool(result.value)
    
    def isWritable(self, feature: str) -> bool:
        """
        Check if the given feature is currently writable.

        See the :data:`~andor3.constants.FEATURES` dictionary for possible ``feature`` values.
        
        :param feature: String describing the camera feature.
        :returns: ``True`` if feature is writable, or ``False`` otherwise.
        """
        result = c_int()
        error = self._dll.AT_IsWritable(self.camera_handle, feature, byref(result))
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
        return bool(result.value)
    
    def isReadOnly(self, feature: str) -> bool:
        """
        Check if the given feature is read-only.

        See the :data:`~andor3.constants.FEATURES` dictionary for possible ``feature`` values.
        
        :param feature: String describing the camera feature.
        :returns: ``True`` if feature is read-only, or ``False`` otherwise.
        """
        result = c_int()
        error = self._dll.AT_IsReadOnly(self.camera_handle, feature, byref(result))
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
        return bool(result.value)

    def setInt(self, feature: str, value: int) -> None:
        """
        Set the value for a given integer feature.

        See the :data:`~andor3.constants.FEATURES` dictionary for possible ``feature`` values.
        
        :param feature: String describing the camera feature.
        :param value: New value for the feature.
        """
        error = self._dll.AT_SetInt(self.camera_handle, feature, c_longlong(value))
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
    
    def getInt(self, feature: str) -> int:
        """
        Get the value for a given integer feature.

        See the :data:`~andor3.constants.FEATURES` dictionary for possible ``feature`` values.
        
        :param feature: String describing the camera feature.
        :returns: Value for the feature.
        """
        result = c_longlong()
        error = self._dll.AT_GetInt(self.camera_handle, feature, byref(result))
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
        return result.value
    
    def getIntMin(self, feature: str) -> int:
        """
        Get the minimum allowed value for a given integer feature.

        See the :data:`~andor3.constants.FEATURES` dictionary for possible ``feature`` values.
        
        :param feature: String describing the camera feature.
        :returns: Minimum allowed value for the feature.
        """
        result = c_longlong()
        error = self._dll.AT_GetIntMin(self.camera_handle, feature, byref(result))
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
        return result.value

    def getIntMax(self, feature: str) -> int:
        """
        Get the maximum allowed value for a given integer feature.

        See the :data:`~andor3.constants.FEATURES` dictionary for possible ``feature`` values.
        
        :param feature: String describing the camera feature.
        :returns: Maximum allowed value for the feature.
        """
        result = c_longlong()
        error = self._dll.AT_GetIntMax(self.camera_handle, feature, byref(result))
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
        return result.value

    def setFloat(self, feature: str, value: float) -> None:
        """
        Set the value for a given floating point number feature.

        See the :data:`~andor3.constants.FEATURES` dictionary for possible ``feature`` values.
        
        :param feature: String describing the camera feature.
        :param value: New value for the feature.
        """
        error = self._dll.AT_SetFloat(self.camera_handle, feature, c_double(value))
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
    
    def getFloat(self, feature: str) -> float:
        """
        Get the value for a given floating point number feature.

        See the :data:`~andor3.constants.FEATURES` dictionary for possible ``feature`` values.
        
        :param feature: String describing the camera feature.
        :returns: Value for the feature.
        """
        result = c_double()
        error = self._dll.AT_GetFloat(self.camera_handle, feature, byref(result))
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
        return result.value
    
    def getFloatMin(self, feature: str) -> float:
        """
        Get the minimum allowed value for a given floating point number feature.

        See the :data:`~andor3.constants.FEATURES` dictionary for possible ``feature`` values.
        
        :param feature: String describing the camera feature.
        :returns: Minimum allowed value for the feature.
        """
        result = c_double()
        error = self._dll.AT_GetFloatMin(self.camera_handle, feature, byref(result))
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
        return result.value

    def getFloatMax(self, feature: str) -> float:
        """
        Get the maximum allowed value for a given floating point number feature.

        See the :data:`~andor3.constants.FEATURES` dictionary for possible ``feature`` values.
        
        :param feature: String describing the camera feature.
        :returns: Maximum allowed value for the feature.
        """
        result = c_double()
        error = self._dll.AT_GetFloatMax(self.camera_handle, feature, byref(result))
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
        return result.value

    def setBool(self, feature: str, value: bool) -> None:
        """
        Set the value for a given boolean feature.

        See the :data:`~andor3.constants.FEATURES` dictionary for possible ``feature`` values.
        
        :param feature: String describing the camera feature.
        :param value: New value for the feature.
        """
        error = self._dll.AT_SetBool(self.camera_handle, feature, c_int(value))
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
    
    def getBool(self, feature: str) -> bool:
        """
        Get the value for a given boolean feature.

        See the :data:`~andor3.constants.FEATURES` dictionary for possible ``feature`` values.
        
        :param feature: String describing the camera feature.
        :returns: Value for the feature.
        """
        result = c_int()
        error = self._dll.AT_GetBool(self.camera_handle, feature, byref(result))
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
        return bool(result.value)
    
    def setEnumIndex(self, feature: str, value: int) -> None:
        """
        Set the value for a given enumerated type feature by its numerical index.

        See the :data:`~andor3.constants.FEATURES` dictionary for possible ``feature`` values.

        Use :meth:`getEnumCount` to determine the maximum value allowed for the enumeration index.
        The methods :meth`isEnumIndexImplemented` and :meth:`isEnumIndexAvailable` can be used to check
        if the value is implemented for the current camera model, and whether the camera is currently in a mode
        where the value is able to be selected.
        The string describing the option corresponding to the enumeration index can be obtained using
        :meth:`getEnumStringByIndex`.
        
        :param feature: String describing the camera feature.
        :param value: New value for the feature.
        """
        error = self._dll.AT_SetEnumIndex(self.camera_handle, feature, c_int(value))
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
    
    def setEnumString(self, feature: str, value: str) -> None:
        """
        Set the value for a given enumerated type feature using its string representation.

        See the :data:`~andor3.constants.FEATURES` dictionary for possible ``feature`` values.

        The string describing the option corresponding to an enumeration index can be obtained using
        :meth:`getEnumStringByIndex`.

        :param feature: String describing the camera feature.
        :param value: New value for the feature.
        """
        error = self._dll.AT_SetEnumString(self.camera_handle, feature, value)
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
    
    def getEnumIndex(self, feature: str) -> int:
        """
        Get the numerical index value for a given enumerated type feature.

        See the :data:`~andor3.constants.FEATURES` dictionary for possible ``feature`` values.

        To obtain the value in the form of its description string, use
        ``cam.getEnumStringByIndex(feature, cam.getEnumIndex(feature))``.

        :param feature: String describing the camera feature.
        :returns: Value for the feature.
        """
        result = c_int()
        error = self._dll.AT_GetEnumIndex(self.camera_handle, feature, byref(result))
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
        return result.value
    
    def getEnumCount(self, feature: str) -> int:
        """
        Get the number of possible values available to an enumerated type feature.

        See the :data:`~andor3.constants.FEATURES` dictionary for possible ``feature`` values.

        The methods :meth`isEnumIndexImplemented` and :meth:`isEnumIndexAvailable` can be used to check
        if a value is implemented for the current camera model, and whether the camera is currently in a mode
        where the value is able to be selected.
        The string describing the option corresponding to the enumeration index can be obtained using
        :meth:`getEnumStringByIndex`.

        :param feature: String describing the camera feature.
        :returns: Maximum value for the feature.
        """
        result = c_int()
        error = self._dll.AT_GetEnumCount(self.camera_handle, feature, byref(result))
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
        return result.value
    
    def isEnumIndexAvailable(self, feature: str, index: int) -> bool:
        """
        Determine if a given value for an enumeration type feature is currently available to be selected.

        See the :data:`~andor3.constants.FEATURES` dictionary for possible ``feature`` values.

        Use :meth:`getEnumCount` to determine the maximum value allowed for the enumeration index.
        The method :meth`isEnumIndexImplemented` can be used to check if the value is implemented
        for the current camera model.
        The string describing the option corresponding to the enumeration index can be obtained using
        :meth:`getEnumStringByIndex`.

        :param feature: String describing the camera feature.
        :param index: Index of enumeration for the feature.
        :returns: Boolean describing if the value for the feature is currently available.
        """
        result = c_int()
        error = self._dll.AT_IsEnumIndexAvailable(self.camera_handle, feature, index, byref(result))
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
        return bool(result.value)
    
    def isEnumIndexImplemented(self, feature: str, index: int) -> bool:
        """
        Determine if a given value for an enumeration type feature is implemented for the current camera.

        See the :data:`~andor3.constants.FEATURES` dictionary for possible ``feature`` values.

        The method :meth:`isEnumIndexAvailable` can be used to check if the camera is currently in a mode
        where the value is able to be selected.
        The string describing the option corresponding to the enumeration index can be obtained using
        :meth:`getEnumStringByIndex`.

        :param feature: String describing the camera feature.
        :param index: Index of enumeration for the feature.
        :returns: Boolean describing if the value for the feature is implemented for the camera.
        """
        result = c_int()
        error = self._dll.AT_IsEnumIndexImplemented(self.camera_handle, feature, index, byref(result))
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
        return bool(result.value)
    
    def getEnumStringByIndex(self, feature: str, index: int) -> str:
        """
        Get the string representation corresponding to the enumeration-type feature's numerical index.

        :param feature: String describing the camera feature.
        :param index: Index of enumeration for the feature.
        :returns: String describing the enumeration index.
        """
        result = create_unicode_buffer(STRING_BUFFER_SIZE)
        error = self._dll.AT_GetEnumStringByIndex(self.camera_handle, feature, index, byref(result), STRING_BUFFER_SIZE)
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
        return result.value
    
    def command(self, feature: str) -> None:
        """
        Activate a command type camera feature.

        A command is a camera feature which takes no parameters, for example ``"AcquisitionStart"``, or ``"SoftwareTrigger"``.

        :param feature: String describing the camera feature.
        """
        error = self._dll.AT_Command(self.camera_handle, feature)
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
    
    def setString(self, feature: str, value: str) -> None:
        """
        Set the value for a given string type feature.

        See the :data:`~andor3.constants.FEATURES` dictionary for possible ``feature`` values.

        :param feature: String describing the camera feature.
        :param value: New value for the feature.
        """
        error = self._dll.AT_SetString(self.camera_handle, feature, value)
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
    
    def getString(self, feature: str) -> str:
        """
        Get the value for a given string feature.

        See the :data:`~andor3.constants.FEATURES` dictionary for possible ``feature`` values.
        
        :param feature: String describing the camera feature.
        :returns: Value for the feature.
        """
        result = create_unicode_buffer(STRING_BUFFER_SIZE)
        error = self._dll.AT_GetString(self.camera_handle, feature, byref(result), STRING_BUFFER_SIZE)
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
        return result.value

    def getStringMaxLength(self, feature: str) -> int:
        """
        Get the maximum possible string length for a string-type feature's values.

        :param feature: String describing the camera feature.
        :returns: Maximum string length for feature values.
        """
        result = c_int()
        error = self._dll.AT_GetStringMaxLength(self.camera_handle, feature, byref(result))
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
        return result.value
    
    # I think the Andor library ends up calling the callback function twice when triggered, it's not my fault!
    def registerFeatureCallback(self, feature: str, callback: typing.Callable, args:tuple=(), kwargs:dict={}) -> None:
        """
        Register a method to call when the selected event occurs.

        The Andor library will call the callback immediately so that the callback handler can
        perform any action on the initial value etc.
        However, I think the Andor library ends up calling the callback function twice when
        triggered which is likely a bug.

        :param feature: String describing the feature event.
        :param callback: Method to call when event occurs.
        :param args: Tuple of positional arguments to pass to ``callback``.
        :param kwargs: Dictionary of keyword arguments to pass to ``callback``.
        """
        # Keys for callback dict are the hash of the tuple of (feature, callback)
        # This means can't use the same feature and function with different calling parameters though.
        # Would we ever want to do that? Maybe. Would instead need to make a unique identifier and return it.
        key = hash((feature, callback))
        if key not in self._callbacks:
            # Make a ctypes wrapper function to pass to the Andor library
            cfunc = _CALLBACK_FUNC(self._feature_callback)
            # We could assume that nothing in the Andor libraries would try to dereference our void pointer,
            # then could just pass the key as is. But to be safe, we'll make a real c_longlong of the key and keep it around.
            ckey = c_longlong(key)
            # Add everything to our dict of callbacks
            self._callbacks[key] = (cfunc, ckey, feature, callback, args, kwargs)
            # Register with Andor library.
            # Note that callback will be called immediately once.
            error = self._dll.AT_RegisterFeatureCallback(self.camera_handle, feature, cfunc, byref(ckey))
            if not error == AT_ERR.SUCCESS:
                self._callbacks.pop(key)
                raise AndorError(error)
        else:
            _log.warning(f"Feature callback for \"{feature}\" already registered to the function \"{callback}\"")
        
    def _feature_callback(self, handle, feature, key_pointer):
        # Dereference the pointer to the c_longlong and get it's value
        key = key_pointer.contents.value
        # Grab everything from our callback dict
        _, _, feature, callback, args, kwargs = self._callbacks[key]
        # Could actually check the camera handle is the one we are currently using
        # and that the feature strings match what we expect. They're probably fine...
        # Call the requested python function and pass given parameters
        callback(*args, **kwargs)
        return 0

    def unregisterFeatureCallback(self, feature: str, callback:typing.Callable) -> None:
        """
        Unregister a previously registered callback method.

        The ``callback`` must have been previously registered to the ``feature`` using
        :meth:`registerFeatureCallback`.

        :param feature: String describing the feature event.
        :param callback: Method previously registered for callbacks.
        """
        # Generate dict key
        key = hash((feature, callback))
        if key in self._callbacks:
            # Grab everything from our callback dict and remove the entry
            cfunc, ckey, feature, callback, _, _ = self._callbacks.pop(key)
            # Unregister with Andor library.
            error = self._dll.AT_UnregisterFeatureCallback(self.camera_handle, feature, cfunc, byref(ckey))
            if not error == AT_ERR.SUCCESS:
                raise AndorError(error)
        else:
            _log.warning(f"Feature callback for \"{feature}\" was not registered to the function \"{callback}\"")

    def queueBuffer(self, count:int=1) -> None:
        """
        Prepare a memory buffer for image storage by the Andor SDK3.

        Note that any previously prepared buffers will be destroyed.
        The Buffer sizes are created to match the currently set image size and bit depth, so the
        image settings should be configured prior to buffer creation.

        Multiple buffers may be prepared in a single call by using the ``count`` parameter.

        :param count: Number of buffers to prepare.
        """
        # First, clear any existing buffer
        self.flush()
        self._image_buffer = None
        self._image_properties = None
        # Andor library wants the buffers to be aligned to 8-byte memory locations.
        # I think numpy does this anyway, but can enforce it with some array slicing.
        # Image size includes optional metadata
        imgsize = self.getInt("ImageSizeBytes")        
        # Padding required at end of image to make imgsize a multiple of 8
        postpad = (8 - imgsize%8)%8
        # Allocate a buffer big enough so can offset if needed
        buffer_size = count*(imgsize + postpad) + 8
        _log.debug(f"Allocating {buffer_size} bytes for image buffers.")
        buffer = np.empty(buffer_size, dtype=np.uint8)
        # Offset from start of buffer to enforce 8 byte alignment
        offset = -buffer.ctypes.data%8
        self._image_buffer = buffer[offset:offset + count*(imgsize + postpad)]
        # Queue up the buffer(s) in the Andor library
        for p in range(self._image_buffer.ctypes.data, self._image_buffer.ctypes.data + self._image_buffer.size, imgsize + postpad):
            error = self._dll.AT_QueueBuffer(self.camera_handle, c_void_p(p), imgsize)
            if not error == AT_ERR.SUCCESS:
                self._image_buffer = None
                self._image_properties = None
                raise AndorError(error)
        # Get the current image properties, which can be used for decoding the raw image data later
        self._image_properties = {
            "metadata" : self.MetadataEnable,
            #"metadata_frame_info" : self.MetadataFrameInfo,
            "metadata_timestamp" : self.MetadataTimestamp,
            "metadata_frame" : self.MetadataFrame,
            "encoding" : self.PixelEncoding[1],
            "size" : imgsize,
            "width" : self.AOIWidth,
            "height" : self.AOIHeight,
            "stride" : self.AOIStride
        }

    def waitBuffer(self, timeout:int=INFINITY, copy:bool=False, requeue:bool=False) -> np.ndarray:
        """
        Wait for the Andor SDK3 to fill a previously prepared memory buffer and return the data.

        The ``timeout`` parameter specifies how long to wait (in milliseconds) for the buffer to
        be filled.
        A value of :data:`~andor3.constants.INFINITY` will wait indefinitely, while a value of zero
        will return immediately if there is no buffer currently filled with new image data.

        Setting the ``copy=True`` parameter will make a copy of the data from the memory buffer,
        otherwise only a reference to the buffer memory will be used. Copying is slower (and uses
        more memory), but may be useful if the memory buffer is at risk of being overwritten by
        new data, for example if ``requeue=True`` is set, or :meth:`queueBuffer` is called to
        create a new buffer space.

        Setting the ``requeue=True`` parameter will allow the Andor SDK3 to re-use the buffer space
        in a circular-buffer type arrangement.
        It's may be a good idea to also set ``copy=True`` to ensure the data isn't overwritten by
        the camera before it's used.

        The raw data is returned as a 1-dimensional numpy array of bytes (uint8).

        :param timeout: Timeout (in milliseconds).
        :param copy: Copy the image data instead of returning a reference.
        :param requeue: Re-queue the image buffer for use again.
        :returns: 1-dimensional numpy array of raw image data.
        """
        data_pointer = c_void_p()
        data_size = c_int()
        error = self._dll.AT_WaitBuffer(self.camera_handle, byref(data_pointer), byref(data_size), timeout)
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
        # Compute slice indices
        start_i = data_pointer.value - self._image_buffer.ctypes.data
        stop_i = start_i + data_size.value
        # Get slice from image buffer
        data = self._image_buffer[start_i:stop_i]
        if copy:
            # Copy data out of buffer before it is requeued and the memory potentially overwritten
            data = data.copy()
        # Requeue the same buffer space if requested.
        if requeue:
            error = self._dll.AT_QueueBuffer(self.camera_handle, data_pointer, data_size)
            if not error == AT_ERR.SUCCESS: raise AndorError(error)
        return data
    
    def flush(self):
        """
        Destroy any existing image buffers.
        """
        self._image_buffer = None
        error = self._dll.AT_Flush(self.camera_handle)
        if not error == AT_ERR.SUCCESS: raise AndorError(error)

    def get_feature(self, feature:str, errors="warn"):
        """
        Queries the value of a camera ``feature``, without needing to know the particular data type.

        Internally, calls the appropriate :meth:`getBool`, :meth:`getInt`, :meth:`getFloat`,
        :meth:`getString` or :meth:`getEnumIndex`/:meth:`getEnumStringByIndex` for ``feature`` and returns the value.
        Enum types return a tuple of (index, string).

        See :data:`~andor3.constants.FEATURES` for a list of camera features.

        By default any errors when querying the value are warned about (``errors="warn"``).
        To instead raise exceptions, set ``errors="raise"``.
        To silence errors completely, set ``errors="ignore"``.
        If a value is unable to be queried and errors are ignored, ``None`` will be returned.

        :param feature: String matching the camera feature to query.
        :param errors: Action to take on errors, either ``ignore``, ``warn``, or ``raise``.
        :returns: Value of the requested feature.
        """
        result = None
        try:
            if feature not in FEATURES:
                raise RuntimeError(f"Unknown feature '{feature}'.")
            if FEATURES[feature] in ("Command", "Event"):
                # Commands don't have values.
                # Events, although documented as integer types, are only used to register callbacks.
                raise RuntimeError(f"Attempted to get the value of a '{FEATURES[feature]}' feature type '{feature}'.")
            elif FEATURES[feature] == "Boolean":
                result = self.getBool(feature)
            elif FEATURES[feature] == "Integer":
                result = self.getInt(feature)
            elif FEATURES[feature] == "Floating Point":
                result = self.getFloat(feature)
            elif FEATURES[feature] == "String":
                result = self.getString(feature)
            elif FEATURES[feature] == "Enumerated":
                index = self.getEnumIndex(feature)
                string = self.getEnumStringByIndex(feature, index)
                result = (index, string)
            else:
                # If we get here there's probably a typo in the FEATURES dictionary...
                raise RuntimeError(f"Unknown data type '{FEATURES[feature]}' for the feature '{feature}'.")
        except Exception as ex:
            if errors == "warn":
                _log.warning(f"Unable to get value of '{feature}': {ex}")
            elif errors == "raise":
                raise
        return result

    def set_feature(self, feature:str, value, errors="warn"):
        """
        Sets the value of a camera ``feature``, without needing to know the particular data type.

        Internally, calls the appropriate ``setBool``, ``setInt``, ``setFloat``, ``setString`` or ``setEnum`` for ``feature``.
        Enum types may be set using either the index or string representation.

        See :data:`~andor3.constants.FEATURES` for a list of camera features.

        By default any errors when setting the value are warned about (``errors="warn"``).
        To instead raise exceptions, set ``errors="raise"``.
        To silence errors completely, set ``errors="ignore"``.
        If a value is unable to be set and errors are ignored, ``None`` will be returned.

        :param feature: String matching the camera feature to set.
        :param value: New value of the feature to set.
        :param errors: Action to take on errors, either ``ignore``, ``warn``, or ``raise``.
        """
        try:
            if feature not in FEATURES:
                raise RuntimeError(f"Unknown feature '{feature}'.")
            if FEATURES[feature] == "Command":
                # Setting a command isn't really correct, but we'll allow it (ignoring the given value)
                self.command(feature)
            elif FEATURES[feature] == "Event":
                # Can't set a value for a feature.
                raise RuntimeError(f"Attempted to set the value of a '{FEATURES[feature]}' feature type '{feature}'.")
            elif FEATURES[feature] == "Boolean":
                self.setBool(feature, bool(value))
            elif FEATURES[feature] == "Integer":
                self.setInt(feature, int(value))
            elif FEATURES[feature] == "Floating Point":
                self.setFloat(feature, float(value))
            elif FEATURES[feature] == "String":
                self.setString(feature, str(value))
            elif FEATURES[feature] == "Enumerated":
                if type(value) == int:
                    self.setEnumIndex(feature, value)
                else:
                    self.setEnumString(feature, str(value))
        except Exception as ex:
            if errors == "warn":
                _log.warning(f"Unable to set value of '{feature}': {ex}")
            elif errors == "raise":
                raise

    def __getattr__(self, name):
        """
        Intercept ``__getattr__`` calls for class properties which match valid camera features.

        Allows the use of feature names as class properties, for example ``t = cam.ExposureTime``.
        """
        if name in FEATURES:
            return self.get_feature(name)
        else:
            raise AttributeError(f"Attempt to get invalid property '{name}' for Andor3.")

    def __setattr__(self, name, value):
        """
        Intercept __setattr__ calls for class properties which match valid camera features.

        Allows the use of feature names as class properties, for example ``cam.ExposureTime = 0.1``.
        """
        if name in FEATURES:
            self.set_feature(name, value)
        else:
            super().__setattr__(name, value)

    def start(self):
        """
        A shortcut for calling ``cam.command("AcquisitionStart")``.
        """
        error = self._dll.AT_Command(self.camera_handle, "AcquisitionStart")
        if not error == AT_ERR.SUCCESS: raise AndorError(error)
    
    def stop(self):
        """
        A shortcut for calling ``cam.command("AcquisitionStop")``.

        Any errors will be ignored (for example, if the camera is already stopped).
        """
        _ = self._dll.AT_Command(self.camera_handle, "AcquisitionStop")

    def min(self, feature:str, errors="warn"):
        """
        Queries the minimum allowed value of a camera ``feature``, without needing to know the particular data type.

        Internally, calls the appropriate ``getIntMin`` or ``getFloatMin`` for ``feature`` and returns the value.

        See :data:`~andor3.constants.FEATURES` for a list of camera features.

        By default any errors when querying the value are warned about (``errors="warn"``).
        To instead raise exceptions, set ``errors="raise"``.
        To silence errors completely, set ``errors="ignore"``.
        If a value is unable to be queried and errors are ignored, ``None`` will be returned.

        :param feature: String matching the camera feature to query.
        :param errors: Action to take on errors, either ``ignore``, ``warn``, or ``raise``.
        :returns: Minimum allowed value of the requested feature.
        """
        result = None
        try:
            if feature not in FEATURES:
                raise RuntimeError(f"Unknown feature '{feature}'.")
            if FEATURES[feature] in ("Command", "Event", "Boolean", "String", "Enumerated"):
                # These don't have minimum values.
                raise RuntimeError(f"Attempted to get the minimum of a '{FEATURES[feature]}' feature type '{feature}'.")
            elif FEATURES[feature] == "Integer":
                result = self.getIntMin(feature)
            elif FEATURES[feature] == "Floating Point":
                result = self.getFloatMin(feature)
            else:
                # If we get here there's probably a typo in the FEATURES dictionary...
                raise RuntimeError(f"Unknown data type '{FEATURES[feature]}' for the feature '{feature}'.")
        except Exception as ex:
            if errors == "warn":
                _log.warning(f"Unable to get minimum of '{feature}': {ex}")
            elif errors == "raise":
                raise
        return result

    def max(self, feature:str, errors="warn"):
        """
        Queries the maximum allowed value of a camera `feature`, without needing to know the particular data type.

        Internally, calls the appropriate ``getIntMax`` or ``getFloatMax`` for ``feature`` and returns the value.

        See :data:`~andor3.constants.FEATURES` for a list of camera features.

        By default any errors when querying the value are warned about (``errors="warn"``).
        To instead raise exceptions, set ``errors="raise"``.
        To silence errors completely, set ``errors="ignore"``.
        If a value is unable to be queried and errors are ignored, ``None`` will be returned.

        :param feature: String matching the camera feature to query.
        :param errors: Action to take on errors, either ``ignore``, ``warn``, or ``raise``.
        :returns: Maximum allowed value of the requested feature.
        """
        result = None
        try:
            if feature not in FEATURES:
                raise RuntimeError(f"Unknown feature '{feature}'.")
            if FEATURES[feature] in ("Command", "Event", "Boolean", "String", "Enumerated"):
                # These don't have maximum values.
                raise RuntimeError(f"Attempted to get the maximum of a '{FEATURES[feature]}' feature type '{feature}'.")
            elif FEATURES[feature] == "Integer":
                result = self.getIntMax(feature)
            elif FEATURES[feature] == "Floating Point":
                result = self.getFloatMax(feature)
            else:
                # If we get here there's probably a typo in the FEATURES dictionary...
                raise RuntimeError(f"Unknown data type '{FEATURES[feature]}' for the feature '{feature}'.")
        except Exception as ex:
            if errors == "warn":
                _log.warning(f"Unable to get maximum of '{feature}': {ex}")
            elif errors == "raise":
                raise
        return result

    @property
    def features(self):
        """
        A list of feature names available to the currently opened camera.
        """
        return [f for f in FEATURES if self.isImplemented(f)]        

    def describe_features(self):
        """
        Generate a string representation of all available camera features and their values.
        """
        s = ""
        for feature, datatype in FEATURES.items():
            if self.isImplemented(feature):
                s += (f"{feature}: "
                    f"{datatype} "
                    f"({'r' if self.isReadable(feature) else '-'}"
                    f"{'w' if self.isWritable(feature) else '-'})\n") 
                if datatype in ("Integer", "Floating Point"):
                    s += (f"  value={self.get_feature(feature):g} "
                        f"min={self.min(feature):g} "
                        f"max={self.max(feature):g}\n")
                elif datatype == "Boolean":
                    s += (f"  value={self.get_feature(feature)}\n")
                elif datatype == "String":
                    s += (f"  value=\"{self.get_feature(feature)}\"\n")
                elif datatype == "Enumerated":
                    for i in range(self.getEnumCount(feature)):
                        if self.isEnumIndexImplemented(feature, i):
                            selected_mark = "->" if self.getEnumIndex(feature) == i else "  "
                            available_mark = ":" if self.isEnumIndexAvailable(feature, i) else "x"
                            s += (f"  {selected_mark} {i:2} {available_mark} {self.getEnumStringByIndex(feature, i)}\n")
        return s.rstrip("\n")

    def decode_image(self, data_raw):
        """
        Decode an image from the raw data byte stream.

        A tuple of the decoded image and timestamp will be returned.
        If metadata is not present, the timestamp will be zero.

        :param data_raw: Raw image byte data, with optional metadata.
        :returns: Tuple of image and timestamp.
        """
        timestamp = 0
        if self._image_properties["metadata"] and self._image_properties["metadata_timestamp"]:
            # Extract timestamp from metadata
            md, data_raw = decode_metadata(data_raw)
            if "timestamp" in md:
                timestamp = md["timestamp"]
        return decode_image_data(
            data_raw=data_raw,
            encoding=self._image_properties["encoding"],
            width=self._image_properties["width"],
            height=self._image_properties["height"],
            stride=self._image_properties["stride"]), timestamp



# Copyright 2021 Patrick C. Tapping
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Utility methods for image data processing.

If the (optional) `numba <http://numba.pydata.org/>`__ library is installed, some methods are able
to be accelerated using the numba just-in-time (JIT) compiler.
In the absence of the library, alternative numpy-based implementations are used.
The ``_np_`` and ``_nb_`` prefixes denote the numpy or numba implementations, respectively.
These should not be called directly, instead, use the non-prefixed version of the methods
documented here which will choose the appropriate implementation for you.
"""

__all__ = ["unpack_uint12", "fvb", "fvb_images", "decode_image_data", "decode_image_with_metadata", "decode_metadata", "FrameServer", "FrameDump"]

import logging
from threading import Thread, Event
import struct
from time import monotonic, sleep

import numpy as np

from . error import *

_log = logging.getLogger(__name__)
"""Logging output for use by this module."""


# numpy routines for data processing, which will be used if the numba library is not available.
# These are generally slower than the numba compiled versions, but not too slow to use.
def _np_unpack_uint12(packed_data):
    """
    Unpacks the 12BitPacked image format into an array of unsigned 16-bit integers.

    `packed_data` is numpy array of ``np.uint8``, consisting of pairs of 12-bit numbers packed into 3 byte sequences.

    :param packed_data: Numpy array of ``np.uint8`` containing 12BitPacked image data.
    :returns: Image data as a numpy array of ``np.uint16``.
    """

    # This routine is written using numpy array operations, which is only about 5x slower than the compiled numba version.
    # This method generally shouldn't be called directly.
    # Instead, use unpack_uint12 which will select the numba or numpy versions as appropriate.

    # packed_data must be at least two values packed into 3 bytes
    if not packed_data.shape[0]%3 == 0:
        raise RuntimeError("Input data size must be a multiple of 3 bytes (pairs of 12-bit numbers packed into 3 bytes)")

    # Output array, 16-bit unsigned integers. Upper 4 bits will be zero.
    out = np.empty(2*(packed_data.shape[0]//3), dtype=np.uint16)

    # This looks simple but uses extra temporary data structures
    #out[0::2] = (packed_data[0::3].astype(np.uint16) << 4) + (packed_data[1::3].astype(np.uint16) & 0xF)
    #out[1::2] = (packed_data[2::3].astype(np.uint16) << 4) + (packed_data[1::3].astype(np.uint16) >> 4)

    # This looks uglier, but is 30% faster
    out[0::2] = packed_data[0::3]
    out[0::2] = out[0::2] << 4
    out[0::2] += packed_data[1::3] & 0xF

    out[1::2] = packed_data[2::3]
    out[1::2] = out[1::2] << 4
    out[1::2] += packed_data[1::3] >> 4

    return out
# Use numpy version by default
unpack_uint12 = _np_unpack_uint12


def _np_fvb(image):
    """
    Perform full-vertical-binning (FVB) of image data.

    ``image`` is a 2-dimensional numpy array of image data, where the first dimension indexes the
    column (width, x-coordinate) of the image, the second the row (height, y-coordinate).
    
    The vertically binned data will be returned as the mean as a 1-dimensional array of float32.

    :param image: Numpy array of image data.
    :returns: Vertically binned array of ``np.float32``.
    """

    # This numpy routine simply takes the mean across the second axis and is fast, but only single-threaded.
    # The numba version is faster due to parallelisation.
    # This method generally shouldn't be called directly.
    # Instead, use fvb which will select the numba or numpy versions as appropriate.

    return np.mean(image, axis=1).astype(np.float32)
# Use numpy version by default
fvb = _np_fvb

def _np_fvb_images(image):
    """
    Perform full-vertical-binning (FVB) of a set of image data.

    ``images`` is a 3-dimensional numpy array of image data in uint16 or uint32 format, where the
    first dimension indexes the image number, the second the column (width, x-coordinate) of the image, the third the row (height, y-coordinate).

    The vertically binned data will be returned as the mean as a 2-dimensional array of float32.

    :param image: Numpy array of image data in uint16 or uint32 format.
    :returns: Vertically binned array of ``np.float32``.
    """
    return np.mean(image, axis=2).astype(np.float32)
# Use numpy version by default
fvb_images = _np_fvb_images


# Use numba accelerated routines if available
try:
    import numba as nb
    
    @nb.njit(nb.uint16[::1](nb.uint8[::1]), fastmath=True, parallel=True)
    def _nb_unpack_uint12(packed_data):
        """
        Unpacks the 12BitPacked image format into an array of unsigned 16-bit integers.

        `packed_data` is numpy array of ``np.uint8``, consisting of pairs of 12-bit numbers packed into 3 byte sequences.

        :param packed_data: Numpy array of ``np.uint8`` containing 12BitPacked image data.
        :returns: Image data as a numpy array of ``np.uint16``.
        """

        # This routine is written in a very simplistic manner which the numba just-in-time (JIT) compiler is easily able to optimize.
        # Running this as pure python code is not advisable! Tests indicate it is about 6500x slower than the compiled version.
        # In case the numba library is unavailable, a numpy routine will be used instead, which is only about 5x slower.
        # This method generally shouldn't be called directly.
        # Instead, use unpack_uint12 which will select the numba or numpy versions as appropriate.

        # packed_data must be at least two values packed into 3 bytes
        if not packed_data.shape[0]%3 == 0:
            raise RuntimeError("Input data size must be a multiple of 3 bytes (pairs of 12-bit numbers packed into 3 bytes)")

        # Output array, 16-bit unsigned integers. Upper 4 bits will be zero.
        out = np.empty(2*(packed_data.shape[0]//3), dtype=np.uint16)

        for i in nb.prange(packed_data.shape[0]//3):
            byte1 = np.uint16(packed_data[i*3])
            byte2 = np.uint16(packed_data[i*3 + 1])
            byte3 = np.uint16(packed_data[i*3 + 2])
            out[2*i]     = (byte1 << 4) + (byte2 & 0xF)
            out[2*i + 1] = (byte3 << 4) + (byte2 >> 4)

        return out
    # Override with numba version
    unpack_uint12 = _nb_unpack_uint12


    @nb.njit([nb.float32[:](nb.uint16[:,:]), nb.float32[:](nb.uint32[:,:])], parallel=True)
    def _nb_fvb(image):
        """
        Perform full-vertical-binning (FVB) of image data.

        ``image`` is a 2-dimensional numpy array of image data in uint16 or uint32 format, where the
        first dimension indexes the column (width, x-coordinate) of the image, the second the row (height, y-coordinate).
        
        The vertically binned data will be returned as the mean as a 1-dimensional array of float32.

        :param image: Numpy array of image data in uint16 or uint32 format.
        :returns: Vertically binned array of ``np.float32``.
        """

        # This numba accelerated routine is not intrinsically faster than the numpy version on a single CPU core,
        # but is faster because of parallelisation.
        # This method generally shouldn't be called directly.
        # Instead, use :meth:`fvb` which will select the numba or numpy versions as appropriate.

        result = np.zeros((image.shape[0],), dtype=np.uint64)
        for col in nb.prange(image.shape[0]):
            for row in nb.prange(image.shape[1]):
                result[col] += image[col,row]
        return (result/image.shape[1]).astype(np.float32)
    # Override with numba version
    fvb = _nb_fvb

    @nb.njit([nb.float32[:,:](nb.uint16[:,:,:]), nb.float32[:,:](nb.uint32[:,:,:])], parallel=True)
    def _nb_fvb_images(images):
        """
        Perform full-vertical-binning (FVB) of a set of image data.

        ``images`` is a 3-dimensional numpy array of image data in uint16 or uint32 format, where the
        first dimension indexes the image number, the second the column (width, x-coordinate) of the image, the third the row (height, y-coordinate).

        The vertically binned data will be returned as the mean as a 2-dimensional array of float32.

        :param image: Numpy array of image data in uint16 or uint32 format.
        :returns: Vertically binned array of ``np.float32``.
        """
        # This tends to outperform the numpy version in certain circumstances
        # It would typically be used when n is large.
        result = np.zeros((images.shape[0], images.shape[1]), dtype=np.uint64)
        for n in nb.prange(images.shape[0]):
            for col in nb.prange(images.shape[1]):
                for row in nb.prange(images.shape[2]):
                    result[n,col] += images[n,col,row]
        return (result/images.shape[2]).astype(np.float32)
    # Override with numba version
    fvb_images = _nb_fvb_images

except:
    _log.warning("The numba library is not available. Pure numpy routines will be used instead but performance may suffer.")


def decode_image_data(data_raw, encoding, width, height, stride):
    """
    Decode raw bytes into an image.

    The decoding process needs to know the ``encoding`` of the data, which should be one of
    ``"Mono12"``, ``"Mono12Packed"``, ``"Mono16"``, or ``"Mono32"``.

    The ``height`` and ``width`` parameters determine the shape of the returned image data in
    pixels. There may be redundant padding bytes at the end of rows of pixels, in which case the
    ``stride`` parameter (in bytes) may be larger than expected given the width and bit-depth of the
    image. Additionally, padding at the end of the image data may be optionally be present, which is
    used by the Andor subsystem to align the data to 8-byte boundaries.

    If metadata is enabled and present in ``data_raw`` it will be ignored. To preserve information
    in the metadata, see the :meth:`decode_image_with_metadata` function.

    :param data_raw: Raw image byte data.
    :param encoding: String describing the image encoding method.
    :param width: Width of the resulting image, in pixels.
    :param height: Height of the resulting image, in pixels.
    :param stride: Number of bytes used to encode a single row of pixels.

    :returns: Image as 2D numpy array with shape (width, height).
    """
    # Expect data_raw to always be uint8 direct from waitBuffer calls.
    if not data_raw.dtype == np.uint8:
        raise RuntimeError("Image decoding requires data_raw to be array of uint8 data type.")

    # Trim off padding and metadata
    data_raw = data_raw[0:stride*height]

    # Convert to correct data type, convert stride from bytes to array elements
    if encoding in ("Mono12", "Mono16"):
        # Interpret data as uint16 type (no copy)
        data = data_raw.view(dtype=np.uint16)
        stride = stride//2
    elif encoding == "Mono12Packed":
        # Unpack Mono12Packed into uint16 type (data copy required)
        data = unpack_uint12(data_raw)
        # Stride so should be 1.5x larger now
        stride = 3*stride//2
    elif encoding == "Mono32":
        # Interpret data as uint32 type (no copy)
        data = data_raw.view(dtype=np.uint32)
        stride = stride//4
    else:
        raise RuntimeError(f"Pixel encoding type '{encoding}' not supported, (must be Mono12, Mono12Packed, or Mono32).")
    
    # Reshape from 1D to a 2D array, will still include padding
    data = data.reshape((stride, height), order="F")
    # Slice data array to remove any padding pixels at end of rows
    data = data[0:width,:]

    return data


def decode_image_with_metadata(data_raw):
    """
    Decode raw bytes into an image, using the included frame info and timestamp metadata.

    A tuple of the decoded image (as 2D numpy array) and timestamp (in clock ticks) will be
    returned.

    Note that this function may be useless, as it is not clear that any camera supports the
    ``MetadataFrameInfo`` property, and thus there may never be the metadata required to determine
    the image dimensions etc.

    :param data_raw: Raw byte data containing image and metadata.
    :returns: Image and metadata timestamp.
    """
    md, imagedata = decode_metadata(data_raw)
    image = decode_image_data(imagedata, md["encoding"], md["width"], md["height"], md["stride"])
    return image, md["timestamp"]


def decode_metadata(metadata_raw):
    """
    Decode a frame metadata block.

    If metadata is enabled for image frames, then the image data contains additional fields appended
    to the pixel data. The information is contained in three different block types:

    - FrameInfo (16 bytes) Height (uint16), width (uint16), [unused] (byte), encoding (uint8),
        stride (uint16), chunk_id=7 (uint32), chunk_length (uint32).
    - Ticks (16 bytes) Ticks (uint64), chunk_id=1 (uint32), chunk_length (uint32).
    - FrameData (variable size) Image (stride × height + padding bytes), chunkid=0 (uint32),
        chunk_length (uint32).

    Values are stored as little-endian. The values of chunk_length do not include the length field
    itself, so are 4-bytes less than the actual chunk size.

    The timestamp is stored as clock ticks since the camera startup. The frequency of the camera
    clock (in Hz) can be obtained using the ``TimeStampClockFrequency`` property.

    This function should be able to handle the image data being missing, in the case where the image
    data has been split from the appended metadata. If the image data is missing, ``None`` will be
    returned in place of the image data.

    The returned dictionary of metadata will have the fields ``height``, ``width``, ``stride``,
    ``encoding`` (as string), and ``timestamp``. The image data will be the raw byte stream as a
    numpy array of size (stride × height + padding bytes). The fields will only be present in the
    dictionary if the relevant blocks have been enabled, for example, buy settings the ``Metadata``
    and ``MetadataTimestamp`` properties to ``True``.

    Note that the ``MetadataFrameInfo`` property and data block format is documented in the metadata
    section of the Andor SDK3 documentation, but is not actually listed as a valid feature for any
    camera model. The Zyla used for testing did not implement ``MetadataFrameInfo``, but the
    decoding of the FrameInfo block should work if a camera happens to support it.

    :param metadata_raw: Image data containing metadata.
    :returns: Tuple of dictionary of metadata and bytes of raw image data.
    """
    metadata = {}
    imagedata = None
    # Start at end of byte stream and work backwards
    p = len(metadata_raw)
    while p >= 8:
        # Read chunk id and length
        p -= 8
        chunk_id, chunk_length = struct.unpack_from("<LL", metadata_raw, offset=p)
        if chunk_id not in (0, 1, 7):
            raise RuntimeError(f"Chunk id={chunk_id} not recognised")
        # Read chunk data block
        p -= chunk_length - 4
        if p < 0:
            # Pointer has gone beyond the start of the byte stream
            if chunk_id == 0:
                # Allow for the image data to be empty, still decode metadata if present
                break
            raise RuntimeError(f"Not enough data to decode chunk id={chunk_id}, length={chunk_length}")
        if chunk_id == 7:
            # FrameInfo chunk
            if not chunk_length == 12:
                raise RuntimeError("FrameInfo chunk (id=7) must have length=12")
            metadata["height"], metadata["width"], encoding, metadata["stride"] = struct.unpack_from("<HHxBH", metadata_raw, offset=p)
            if 0 <= encoding <= 3:
                metadata["encoding"] = ["Mono16", "Mono12", "Mono12Packed", "Mono32"][encoding]
            else:
                metadata["encoding"] = "Unknown"
        elif chunk_id == 1:
            # Ticks chunk
            if not chunk_length == 12:
                raise RuntimeError("Ticks chunk (id=1) must have length=12")
            metadata["timestamp"] = struct.unpack_from("<Q", metadata_raw, offset=p)[0]
        elif chunk_id == 0:
            # FrameData chunk
            if imagedata:
                raise RuntimeError("Duplicate FrameData chunk found")
            if chunk_length < 1:
                raise RuntimeError("FrameData chunk has zero length")
            imagedata = np.array(metadata_raw[p:p + chunk_length - 4], dtype=np.uint8)
            # still raw bytes, could contain padding
    return metadata, imagedata
            

class FrameServer():
    """
    Class which can start an acquisition, then serve each frame to a given function.

    The frame callback method should take the form of ``frame_callback(n, data, timestamp)``,
    where ``n`` is the frame number (zero-based) in the acquisition series, ``data`` is the
    image data, and timestamp is the metadata timestamp if metadata is enabled, else zero.
    
    The optional completion callback method should take the form of ``completion_callback(n)``,
    where ``n`` is the number of frames which were acquired.

    If ``fvb=True``, then the data provided to the callback method is a 1-dimensional numpy array.
    Otherwise, the data is a 2-dimensional numpy array with axes (row, column).

    The rate that frames are served can be limited using the ``frame_rate_max`` parameter.
    A value of zero or ``None`` will not restrict the frame rate. Note that this is the rate
    of frames served, not the actual acquisition rate.

    :param cam: :class:`~andor3.andor3.Andor3` camera to use for frame acquisition.
    :param frame_callback: Function to call on each frame acquisition event.
    :param completion_callback: Function to call when acquisition is completed.
    :param fvb: Perform full-vertical-binning on the image data.
    :param frame_rate_max: Maximum frame serving rate in frames per second.
    """
    def __init__(self, cam, frame_callback, completion_callback=None, fvb=False, frame_rate_max=None):
        self._cam = cam
        self._thread = None
        self._thread_stop = Event()
        if callable(frame_callback):
            self._frame_callback = frame_callback
        else:
            raise TypeError("FrameServer completion_callback parameter is not a callable function.")
        if callable(completion_callback):
            self._completion_callback = completion_callback
        else:
            if completion_callback is not None:
                _log.warning("FrameServer completion_callback parameter is not a callable function, ignoring.")
            self._completion_callback = None
        self._fvb = bool(fvb)
        self._frame_interval = 1.0/frame_rate_max if frame_rate_max else 0.0

    def start(self, nbuffers=10, fvb=None, frame_rate_max=None):
        """
        Start the camera and begin serving acquired frames.

        A circular series of ``nbuffers`` buffers are used, and the returned image data may be only
        a view of the raw camera buffer memory (not copied). If the image data is to be retained for
        long periods, it should be copied into its own memory, either indirectly by a mathematical
        operation or explicitly by something like ``img = data.copy()``. If the ``fvb=True`` option
        is used, the served data is computed from the buffer and so this is not an issue.

        Full-vertical binning can be enabled by setting ``fvb=True``.

        The rate that frames are served can be limited using the ``frame_rate_max`` parameter.
        A value of zero or ``None`` will not restrict the frame rate. Note that this is the rate
        of frames served, not the actual acquisition rate.

        :param nbuffers: Number of frames in the circular image buffer.
        :param fvb: Perform full-vertical-binning on the image data.
        :param frame_rate_max: Maximum acquisition rate in frames per second.
        """
        self._cam.queueBuffer(nbuffers)
        if fvb is not None:
            self._fvb = bool(fvb)
        if frame_rate_max is not None:
            self._frame_interval = 1.0/frame_rate_max if frame_rate_max else 0.0
        self._thread_stop.clear()
        self._thread = Thread(target=self._run, name="Andor3_FrameServer", daemon=True)
        self._thread.start()

    def stop(self):
        """
        Stop the FrameServer.
        """
        self._thread_stop.set()
        # Wait for thread to finish
        if self._thread and self._thread.is_alive():
            self._thread.join()

    def _run(self):
        """
        Run the acquisition thread.
        """
        frame_count = 0
        self._cam.CycleMode = "Continuous"
        last_frame_time = 0.0
        self._cam.start()
        while (not self._thread_stop.is_set()) and self._cam.CameraAcquiring:
            # Requeue to use as a circular buffer
            try:
                data = self._cam.waitBuffer(timeout=200, copy=False, requeue=True)
            except AndorError as ex:
                if ex.code == AT_ERR.TIMEDOUT:
                    # Ignore timeout errors
                    continue
                elif ex.code == AT_ERR.NODATA:
                    # Something nondescript went wrong
                    break
                else:
                    raise
            # Rate limit the frame serving if a maximum frame rate was specified
            if self._frame_interval:
                this_frame_time = monotonic()
                if this_frame_time < last_frame_time + self._frame_interval:
                    continue
            # Notify listener a frame has been acquired
            img, timestamp = self._cam.decode_image(data)
            if self._fvb:
                img = fvb(img)
            self._frame_callback(frame_count, img, timestamp)
            frame_count += 1
            if self._frame_interval:
                last_frame_time = monotonic()
        # Stop the acquisition if it isn't already
        self._cam.stop()
        # Notify listener the acquisition is completed
        if self._completion_callback is not None:
            self._completion_callback(frame_count)
        self._thread = None


class FrameDump():
    """
    Class which can start an acquisition, then notify on the conclusion of the acquisition event
    with all the acquired data.

    This acquisition style pre-allocates all memory for the frame data and thus can use a large
    amount of RAM. Ensure your system has the available RAM for the number and size of images you
    wish to acquire.

    The callback method should take the form of ``completion_callback(data, timestamps)``,
    where ``data`` is a numpy array of the image data, and timestamps is a numpy array of
    timestamps for the data if metadata is enabled (zeros otherwise).

    If ``fvb=True``, then the data provided to the callback method is a 2-dimensional numpy array,
    with axes being (n, column). Otherwise, the data is a 3-dimensional numpy array with axes (n,
    row, column).

    :param cam: :class:`~andor3.andor3.Andor3` camera to use for frame acquisition.
    :param completion_callback: Function to call when acquisition is completion.
    :param fvb: Perform full-vertical-binning on the image data.
    """

    def __init__(self, cam, completion_callback, fvb=False):
        self._cam = cam
        self._thread = None
        self._thread_stop = Event()
        if callable(completion_callback):
            self._completion_callback = completion_callback
        else:
            raise TypeError("FrameDump completion_callback parameter is not a callable function.")
        self._fvb = bool(fvb)

    def start(self, n_images=None, fvb=None):
        """
        Start the acquisition process.

        A maximum of ``n_images`` frames will be acquired. If set to ``None`` (default), the number
        of images will be determined by the camera's ``FrameCount`` and ``AccumulateCount``
        properties.

        Full-vertical binning can be enabled or disabled using the ``fvb`` parameter.

        :param n_images: Maximum number of images to capture.
        :param fvb: Perform full-vertical-binning on the image data.
        """
        # TODO: Should check here for supported pixel format etc
        self._cam.CycleMode = "Fixed"
        if n_images:
            self._cam.FrameCount = int(n_images)*self._cam.AccumulateCount
        else:
            n_images = self._cam.FrameCount//self._cam.AccumulateCount
        self._cam.queueBuffer(n_images)
        if fvb is not None:
            self._fvb = bool(fvb)
        self._thread_stop.clear()
        self._thread = Thread(target=self._run, args=(n_images,), name="Andor3_FrameDump", daemon=True)
        self._thread.start()

    def stop(self):
        """
        Stop the camera.
        """
        self._thread_stop.set()
        # Wait for thread to finish
        if self._thread and self._thread.is_alive():
            self._thread.join() 

    def _run(self, frame_count_max):
        # Camera does not automatically stop acquiring once FrameCount has been reached???
        frame_count = 0
        self._cam.start()
        while (not self._thread_stop.is_set()) and (frame_count < frame_count_max) and self._cam.CameraAcquiring:
            try:
                # Do we really have to do this? Can't the camera just fill the buffers itself?
                self._cam.waitBuffer(timeout=250)
            except AndorError as ex:
                if ex.code == AT_ERR.TIMEDOUT:
                    # Ignore timeout errors
                    continue
                elif ex.code == AT_ERR.NODATA:
                    # Something nondescript went wrong
                    break
                else:
                    raise
            frame_count += 1
        # Stop the acquisition if it isn't already
        self._cam.stop()
        
        # Process the raw byte data to extract the images
        encoding = self._cam._image_properties["encoding"] # string
        imgsize = self._cam._image_properties["size"] # bytes, including metadata
        postpad = (8 - imgsize%8)%8                   # bytes, for 8-byte alignment
        w = self._cam._image_properties["width"]      # pixels
        s = self._cam._image_properties["stride"]     # bytes
        h = self._cam._image_properties["height"]     # pixels
        count = self._cam._image_buffer.shape[0]//imgsize
        n = min(count, frame_count)                    
        # View the buffer as separate byte blocks for each image
        buffer = self._cam._image_buffer.reshape((count, imgsize + postpad))[:,:imgsize]
        # Metadata may be present at end of each image data block
        metadata_raw = buffer[:,s*h:]
        timestamps = np.zeros(metadata_raw.shape[0], dtype=np.uint64)
        if metadata_raw.shape[1] > 0:
            try:
                for i, md in enumerate(metadata_raw):
                    timestamps[i] = decode_metadata(md)[0]["timestamp"]
            except: pass
        if encoding == "Mono12Packed":
            # For 12-bit images, we have to allocate a larger buffer to unpack into
            # This means a (temporary) huge increase in RAM use...
            images = np.empty((count, (3*s//2)*h), dtype=np.uint16)
            for i in range(images.shape[0]):
                images[i] = unpack_uint12(buffer[i])
            # Updated stride value (in bytes) is 1.5x larger now
            s = 3*s//2
            images = images.reshape((count, h, s)).view(dtype=np.uint16)[:n,:,:w]
        else:
            # For 16 or 32 bit images, we should be able to decode image data in place
            # Reinterpret buffer as new data type, crop off metadata, reshape to image dimensions
            if encoding == "Mono32":
                images = buffer.view(dtype=np.uint32)[:,:h*s//4].reshape((count, h, s//4))[:n,:,:w]
            else:
                images = buffer.view(dtype=np.uint16)[:,:h*s//2].reshape((count, h, s//2))[:n,:,:w] 
        # Swap axes to (count, width, height)
        images = np.swapaxes(images, 1, 2)
        # Perform vertical binning if requested
        # TODO: There's probably a less RAM-hungry method to do this
        if self._fvb:
            images = fvb_images(images)
        # The buffer may still be referenced (as "images"), but we still want to get it garbage collected ASAP...
        self._cam.flush()
        # Notify that acquisition has completed
        self._completion_callback(images, timestamps)
        self._thread = None 
     





# Copyright 2021 Patrick C. Tapping
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Definitions of special values, feature strings, and camera handles used by the Andor SDK.
"""

__all__ = ["STRING_BUFFER_SIZE", "INFINITY", "FEATURES", "AT_HANDLE"]

from enum import IntEnum


# Maximum size of strings sent/returned
STRING_BUFFER_SIZE = 1024
"""Buffer size to use when sending or receiving strings from the camera."""

# Infinite timeout when waiting on image buffer
INFINITY = 0xFFFFFFFF
"""Value to indicate an infinite timeout value when waiting on an image buffer."""

# Dictionary of features and their corresponding data type
FEATURES = {
    "AccumulateCount": "Integer",
    "AcquiredCount": "Integer",
    "AcquisitionStart": "Command",
    "AcquisitionStop": "Command",
    "AlternatingReadoutDirection": "Boolean",
    "AOIBinning": "Enumerated",
    "AOIHBin": "Integer",
    "AOIHeight": "Integer",
    "AOILayout": "Enumerated",
    "AOILeft": "Integer",
    "AOIStride": "Integer",
    "AOITop": "Integer",
    "AOIVBin": "Integer",
    "AOIWidth": "Integer",
    "AuxiliaryOutSource": "Enumerated",
    "AuxOutSourceTwo": "Enumerated",
    "BackoffTemperatureOffset": "Floating Point",
    "Baseline": "Integer",
    "BitDepth": "Enumerated",
    "BufferOverflowEvent": "Event",
    "BytesPerPixel": "Floating Point",
    "CameraAcquiring": "Boolean",
    "CameraDump": "Command",
    "CameraFamily": "String",
    "CameraMemory": "Integer",
    "CameraModel": "String",
    "CameraName": "String",
    "CameraPresent": "Boolean",
    "ColourFilter": "Enumerated",
    "ControllerID": "String",
    "CoolerPower": "Floating Point",
    "CycleMode": "Enumerated",
    "DDGIOCEnable": "Boolean",
    "DDGIOCNumberOfPulses": "Integer",
    "DDGIOCPeriod": "Integer",
    "DDGOutputDelay": "Integer",
    "DDGOutputEnable": "Boolean",
    "DDGOutputStepEnable": "Boolean",
    "DDGOpticalWidthEnable": "Boolean",
    "DDGOutputPolarity": "Enumerated",
    "DDGOutputSelector": "Enumerated",
    "DDGOutputWidth": "Integer",
    "DDGStepCount": "Integer",
    "DDGStepDelayCoefficientA": "Floating Point",
    "DDGStepDelayCoefficientB": "Floating Point",
    "DDGStepDelayMode": "Enumerated",
    "DDGStepEnabled": "Boolean",
    "DDGStepUploadProgress": "Integer",
    "DDGStepUploadRequired": "Boolean",
    "DDGStepWidthCoefficientA": "Floating Point",
    "DDGStepWidthCoefficientB": "Floating Point",
    "DDGStepWidthMode": "Enumerated",
    "DDGStepUploadModeValues": "Command",
    "DDR2Type": "String",
    "DeviceCount": "Integer",
    "DeviceVideoIndex": "Integer",
    "DisableShutter": "Boolean",
    "DriverVersion": "String",
    "ElectronicShutteringMode": "Enumerated",
    "EventEnable": "Boolean",
    "EventsMissedEvent": "Event",
    "EventSelector": "Enumerated",
    "ExposedPixelHeight": "Integer",
    "ExposureTime": "Floating Point",
    "ExposureEndEvent": "Event",
    "ExposureStartEvent": "Event",
    "ExternalIOReadout": "Boolean",
    "ExternalTriggerDelay": "Floating Point",
    "FanSpeed": "Enumerated",
    "FastAOIFrameRateEnable": "Boolean",
    "FirmwareVersion": "String",
    "ForceShutterOpen": "Boolean",
    "FrameCount": "Integer",
    "FrameInterval": "Floating Point",
    "FrameIntervalTiming": "Boolean",
    "FrameRate": "Floating Point",
    "FullAOIControl": "Boolean",
    "GateMode": "Enumerated",
    "HeatSinkTemperature": "Floating Point",
    "I2CAddress": "Integer",
    "I2CByte": "Integer",
    "I2CByteCount": "Integer",
    "I2CByteSelector": "Integer",
    "I2CRead": "Command",
    "I2CWrite": "Command",
    "ImageSizeBytes": "Integer",
    "InputVoltage": "Floating Point",
    "InsertionDelay": "Enumerated",
    "InterfaceType": "String",
    "IOControl": "Enumerated",
    "IODirection": "Enumerated",
    "IOState": "Boolean",
    "IOInvert": "Boolean",
    "IOSelector": "Enumerated",
    "IRPreFlashEnable": "Boolean",
    "KeepCleanEnable": "Boolean",
    "KeepCleanPostExposureEnable": "Boolean",
    "LineScanSpeed": "Floating Point",
    "LUTIndex": "Integer",
    "LUTValue": "Integer",
    "MaxInterfaceTransferRate": "Floating Point",
    "MCPGain": "Integer",
    "MCPIntelligate": "Boolean",
    "MCPVoltage": "Integer",
    "MetadataEnable": "Boolean",
    "MetadataFrame": "Boolean",
    "MetadataFrameInfo": "Boolean", # Not in documentation, indicates whether frame info is contained in metadata.
    "MetadataTimestamp": "Boolean",
    "MicrocodeVersion": "String",
    "MultitrackBinned": "Boolean",
    "MultitrackCount": "Integer",
    "MultitrackEnd": "Integer",
    "MultitrackSelector": "Integer",
    "MultitrackStart": "Integer",
    "Overlap": "Boolean",
    "PIVEnable": "Boolean",
    "PixelCorrection": "Boolean", # Documentation says Enumerated on SimCam only, but implemented on Zyla and returns a boolean.
    "PixelEncoding": "Enumerated",
    "PixelHeight": "Floating Point",
    "PixelReadoutRate": "Enumerated",
    "PixelWidth": "Floating Point",
    "PortSelector": "Integer",
    "PreAmpGain": "Enumerated",
    "PreAmpGainChannel": "Enumerated",
    "PreAmpGainControl": "Enumerated",
    "PreAmpGainValue": "Integer",
    "PreAmpGainSelector": "Enumerated",
    "PreAmpOffsetValue": "Integer",
    "PreTriggerEnable": "Boolean",
    "ReadoutTime": "Floating Point",
    "RollingShutterGlobalClear": "Boolean",
    "RowNExposureEndEvent": "Event",
    "RowNExposureStartEvent": "Event",
    "RowReadTime": "Floating Point",
    "ScanSpeedControlEnable": "Boolean",
    "SensorCooling": "Boolean",
    "SensorHeight": "Integer",
    "SensorModel": "String",
    "SensorReadoutMode": "Enumerated",
    "SensorType": "String",  # Documentation says Enumerated on Apogee only, but implemented on Zyla and returns a string.
    "SensorTemperature": "Floating Point",
    "SensorWidth": "Integer",
    "SerialNumber": "String",
    "ShutterAmpControl": "Boolean",
    "ShutterMode": "Enumerated",
    "ShutterOutputMode": "Enumerated",
    "ShutterState": "Boolean",
    "ShutterStrobePeriod": "Floating Point",
    "ShutterStrobePosition": "Floating Point",
    "ShutterTransferTime": "Floating Point",
    "SimplePreAmpGainControl": "Enumerated",
    "SoftwareTrigger": "Command",
    "SoftwareVersion": "String",
    "SpuriousNoiseFilter": "Boolean",
    "StaticBlemishCorrection": "Boolean",
    "SynchronousTriggering": "Boolean",
    "TargetSensorTemperature": "Floating Point",
    "TemperatureControl": "Enumerated",
    "TemperatureStatus": "Enumerated",
    "TimestampClock": "Integer",
    "TimestampClockFrequency": "Integer",
    "TimestampClockReset": "Command",
    "TransmitFrames": "Boolean",
    "TriggerMode": "Enumerated",
    "UsbProductId": "Integer",
    "UsbDeviceId": "Integer",
    "VerticallyCentreAOI": "Boolean"
}
"""
Dictionary containing valid camera feature strings and their corresponding data type.
"""


class AT_HANDLE(IntEnum):
    """
    Enumeration of special camera handle values used by the Andor SDK3.
    """
    UNINITIALISED = -1
    SYSTEM = 1